/* Copyright (c) 1996-98, Timothy Mann */

/* This software may be copied, modified, and used for any purpose
 * without fee, provided that (1) the above copyright notice is
 * retained, and (2) modified versions are clearly marked as having
 * been modified, with the modifier's name and the date included.  */

/* Last modified on Fri Apr 24 21:18:06 PDT 1998 by mann */

/*
 * mkdisk.c
 * Make a blank (unformatted) emulated floppy or hard drive in a file,
 * or write protect/unprotect an existing one.
 */

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

void Usage(char *progname)
{
  fprintf(stderr, 
	  "Usage:\t%s -1 file\n"
	  "\t%s [-3] file\n"
	  "\t%s -k [-s sides] [-d density] [-8] [-i] file\n"
	  "\t%s -h [-c cyl] [-s sec] [-g gran] [-d dir] file\n"
	  "\t%s mkdisk {-p|-u} [-1|-3|-k|-h]\n",
	  progname, progname, progname, progname, progname);
  exit(2);
}

typedef unsigned char Uchar;

/* Matthew Reed's hard drive format.  Thanks to Matthew for providing
   documentation.  The comments below are copied from his mail
   messages, with some additions. */

typedef struct {
  Uchar id1;       /* 0: Identifier #1: 56H */
  Uchar id2;       /* 1: Identifier #2: CBH */
  Uchar ver;       /* 2: Version of format: 10H = version 1.0 */
  Uchar cksum;     /* 3: Simple checksum: 
		      To calculate, add together bytes 0 to 31 of header
		      (excepting byte 3), then XOR result with 4CH */
  Uchar blks;      /* 4: Number of 256 byte blocks in header: should be 1 */
  Uchar mb4;       /* 5: Not used, but HDFORMAT sets to 4 */
  Uchar media;     /* 6: Media type: 0 for hard disk */
  Uchar flag1;     /* 7: Flags #1:
		      bit 7: Write protected: 0 for no, 1 for yes 
                             [xtrs ignores for now]
		      bit 6: Must be 0
		      bit 5 - 0: reserved */
  Uchar flag2;     /* 8: Flags #2: reserved */
  Uchar flag3;     /* 9: Flags #3: reserved */
  Uchar crtr;      /* 10: Created by: 
		      14H = HDFORMAT
		      42H = xtrs mkdisk
                      80H = Cervasio xtrshard port to Vavasour M4 emulator */
  Uchar dfmt;      /* 11: Disk format: 0 = LDOS/LS-DOS */
  Uchar mm;        /* 12: Creation month: mm */
  Uchar dd;        /* 13: Creation day: dd */
  Uchar yy;        /* 14: Creation year: yy (offset from 1900) */
  Uchar res1[12];  /* 15 - 26: reserved */
  Uchar dparm;     /* 27: Disk parameters: (unused with hard drives)
		      bit 7: Density: 0 = double, 1 = single
		      bit 6: Sides: 0 = one side, 1 = 2 sides
		      bit 5: First sector: 0 if sector 0, 1 if sector 1
		      bit 4: DAM convention: 0 if normal (LDOS),
		      1 if reversed (TRSDOS 1.3)
		      bit 3 - 0: reserved */
  Uchar cyl;       /* 28: Number of cylinders per disk */
  Uchar sec;       /* 29: Number of sectors per track (floppy); cyl (hard) */
  Uchar gran;      /* 30: Number of granules per track (floppy); cyl (hard)*/
  Uchar dcyl;      /* 31: Directory cylinder [mkdisk sets to 1; xtrs ignores]*/
  char label[32];  /* 32: Volume label: 31 bytes terminated by 0 */
  char filename[8];/* 64 - 71: 8 characters of filename (without extension)
		      [Cervasio addition.  xtrs actually doesn't limit this 
                       to 8 chars or strip the extension] */
  Uchar res2[184]; /* 72 - 255: reserved */
} ReedHardHeader;

ReedHardHeader rhh;

int
main(int argc, char *argv[])
{
  int jv1 = 0, jv3 = 0, dmk = 0, hard = 0;
  int cyl = -1, sec = -1, gran = -1, dir = -1, eight = 0, ignden = 0;
  int writeprot = 0, unprot = 0;
  int i, c, oumask;
  char *fname;
  FILE *f;

  opterr = 0;
  for (;;) {
    c = getopt(argc, argv, "13khc:s:g:d:8ipu");
    if (c == -1) break;
    switch (c) {
    case '1':
      jv1 = 1;
      break;
    case '3':
      jv3 = 1;
      break;
    case 'k':
      dmk = 1;
      break;
    case 'h':
      hard = 1;
      break;
    case 'c':
      cyl = atoi(optarg);
      break;
    case 's':
      sec = atoi(optarg);
      break;
    case 'g':
      gran = atoi(optarg);
      break;
    case 'd':
      dir = atoi(optarg);
      break;
    case '8':
      eight = 1;
      break;
    case 'i':
      ignden = 1;
      break;
    case 'p':
      writeprot = 1;
      break;
    case 'u':
      unprot = 1;
      break;
    case '?':
    default:
      Usage(argv[0]);
      break;
    }
  }

  if (argc - optind != 1) Usage(argv[0]);

  fname = argv[optind];

  if (writeprot || unprot) {
    /* Completely different functionality here */
    struct stat st;
    int newmode;

    if (writeprot && unprot) {
      fprintf(stderr,
	      "%s: -p and -u are mutually exclusive\n", argv[0]);
      exit(2);
    }    

    if (jv1 + jv3 + dmk + hard != 1) {
      fprintf(stderr,
	      "%s: %s requires exactly one of -1, -3, -k, or -h\n",
	      argv[0], writeprot ? "-p" : "-u");
      exit(2);
    }

    if (stat(fname, &st) < 0) {
      perror(fname);
      exit(1);
    }

    /* Make writable so we can poke inside if need be */
    if (chmod(fname, st.st_mode | (S_IWUSR|S_IWGRP|S_IWOTH)) < 0) {
      perror(fname);
      exit(1);
    }

    f = fopen(fname, "r+");
    if (f == NULL) {
      perror(fname);
      exit(1);
    }

    /* Poke inside */
    if (jv1) {
      /* No indication inside; nothing to do here */

    } else if (jv3) {
      /* Set the magic byte */
      fseek(f, 256*34-1, 0);
      putc(writeprot ? 0 : 0xff, f);

    } else if (dmk) {
      /* Set the magic byte */
      putc(writeprot ? 0xff : 0, f);

    } else if (hard) {
      /* Set the magic bit */
      fseek(f, 7, 0);
      newmode = getc(f);
      if (newmode == EOF) {
	perror(fname);
	exit(1);
      }
      newmode = (newmode & 0x7f) | (writeprot ? 0x80 : 0);
      fseek(f, 7, 0);
      putc(newmode, f);

    }

    /* Finish by chmoding the file appropriately */
    if (writeprot) {
      newmode = st.st_mode & ~(S_IWUSR|S_IWGRP|S_IWOTH);
    } else {
      oumask = umask(0);
      umask(oumask);
      newmode = st.st_mode |
	(( (st.st_mode & (S_IRUSR|S_IRGRP|S_IROTH)) >> 1 ) & ~oumask);
    }
    if (fchmod(fileno(f), newmode)) {
      perror(fname);
      exit(1);
    }
    fclose(f);
    exit(0);
  }

  switch (jv1 + jv3 + dmk + hard) {
  case 0:
    jv3 = 1;
    break;
  case 1:
    break;
  default:
    fprintf(stderr,
	    "%s: -1, -3, -k, and -h are mutually exclusive\n", argv[0]);
    exit(2);
  }

  if ((jv1 || jv3) && (cyl >= 0 || sec >= 0 || gran >= 0 || dir >= 0)) {
    fprintf(stderr,
	    "%s: -c, -s, -g, -d are not meaningful with -1 or -3\n", argv[0]);
    exit(2);
  }

  if (dmk && (cyl >= 0 || gran >= 0)) {
    fprintf(stderr, "%s: -c and -g are not meaningful with -k\n", argv[0]);
    exit(2);
  }

  if (!dmk && (eight || ignden)) {
    fprintf(stderr, "%s: -8 and -i are only meaningful with -k\n", argv[0]);
    exit(2);
  }    

  if (jv1) {
    /* Unformatted JV1 disk - just an empty file! */
    f = fopen(fname, "w");
    if (f == NULL) {
      perror(fname);
      exit(1);
    }

  } else if (jv3) {
    /* Unformatted JV3 disk. */
    f = fopen(fname, "w");
    if (f == NULL) {
      perror(fname);
      exit(1);
    }
    for (i=0; i<(256*34); i++) {
      putc(0xff, f);
    }

  } else if (dmk) {
    /* Unformatted DMK disk */
    /* Reuse flag letters s, d */
#define sides sec
#define density dir
    if (sides == -1) sides = 2;
    if (density == -1) density = 2;

    if (sides != 1 && sides != 2) {
      fprintf(stderr, "%s error: sides must be 1 or 2\n", argv[0]);
      exit(2);
    }	    
    if (density < 1 || density > 2) {
      fprintf(stderr, "%s error: density must be 1 or 2\n", argv[0]);
      exit(2);
    }	    
    
    f = fopen(fname, "w");
    putc(0, f);           /* 0: not write protected */
    putc(0, f);           /* 1: initially zero tracks */
    if (eight) {
      if (density == 1)
	i = 0x14e0;
      else
	i = 0x2940;
    } else {
      if (density == 1)
	i = 0x0cc0;
      else
	i = 0x1900;
    }
    putc(i & 0xff, f);    /* 2: LSB of track length */
    putc(i >> 8, f);      /* 3: MSB of track length */
    i = 0;
    if (sides == 1)   i |= 0x10;
    if (density == 1) i |= 0x40;
    if (ignden)       i |= 0x80;
    putc(i, f);           /* 4: options */
    putc(0, f);           /* 5: reserved */
    putc(0, f);           /* 6: reserved */
    putc(0, f);           /* 7: reserved */
    putc(0, f);           /* 8: reserved */
    putc(0, f);           /* 9: reserved */
    putc(0, f);           /* a: reserved */
    putc(0, f);           /* b: reserved */
    putc(0, f);           /* c: MBZ */
    putc(0, f);           /* d: MBZ */
    putc(0, f);           /* e: MBZ */
    putc(0, f);           /* f: MBZ */

  } else /* hard */ {
    /* Unformatted hard disk */
    /* We don't care about most of this header, but we generate
       it just in case some user wants to exchange hard drives with
       Matthew Reed's emulator or with Pete Cervasio's port of
       xtrshard/dct to Jeff Vavasour's Model III/4 emulator.
    */
    time_t tt = time(0);
    struct tm *lt = localtime(&tt);
    Uchar *rhhp;
    int cksum;

    if (cyl == -1) cyl = 202;
    if (sec == -1) sec = 256;
    if (gran == -1) gran = 8;
    if (dir == -1) dir = 1;

    if (cyl < 3) {
      fprintf(stderr, "%s error: cyl < 3\n", argv[0]);
      exit(2);
    }
    if (cyl > 203) {
      fprintf(stderr, "%s error: cyl > 203\n", argv[0]);
      exit(2);
    }
    if (sec < 4) {
      fprintf(stderr, "%s error: sec < 4\n", argv[0]);
      exit(2);
    }
    if (sec > 256) {
      fprintf(stderr, "%s error: sec > 256\n", argv[0]);
      exit(2);
    }
    if (gran < 1) {
      fprintf(stderr, "%s error: gran < 1\n", argv[0]);
      exit(2);
    }
    if (gran > 8) {
      fprintf(stderr, "%s error: gran > 8\n", argv[0]);
      exit(2);
    }
    if (sec < gran) {
      fprintf(stderr, "%s error: sec < gran\n", argv[0]);
      exit(2);
    }
    if (sec % gran != 0) {
      fprintf(stderr, "%s error: sec %% gran != 0\n", argv[0]);
      exit(2);
    }
    if (sec / gran > 32) {
      fprintf(stderr, "%s error: sec / gran > 32\n", argv[0]);
      exit(2);
    }
    if (dir < 1) {
      fprintf(stderr, "%s error: dir < 1\n", argv[0]);
      exit(2);
    }
    if (dir >= cyl) {
      fprintf(stderr, "%s error: dir >= cyl\n", argv[0]);
      exit(2);
    }

    rhh.id1 = 0x56;
    rhh.id2 = 0xcb;
    rhh.ver = 0x10;
    rhh.cksum = 0;  /* init for cksum computation */
    rhh.blks = 1;
    rhh.mb4 = 4;
    rhh.media = 0;
    rhh.flag1 = 0;
    rhh.flag2 = rhh.flag3 = 0;
    rhh.crtr = 0x42;
    rhh.dfmt = 0;
    rhh.mm = lt->tm_mon + 1;
    rhh.dd = lt->tm_mday;
    rhh.yy = lt->tm_year;
    rhh.dparm = 0;
    rhh.cyl = cyl;
    rhh.sec = sec;
    rhh.gran = gran;
    rhh.dcyl = dir;
    strcpy(rhh.label, "xtrshard");
    strcpy(rhh.filename, fname); /* note we don't limit to 8 chars */

    cksum = 0;
    rhhp = (Uchar *) &rhh;
    for (i=0; i<=31; i++) {
      cksum += rhhp[i];
    }
    rhh.cksum = ((Uchar) cksum) ^ 0x4c;

    f = fopen(fname, "w");
    if (f == NULL) {
      perror(fname);
      exit(1);
    }
    fwrite(&rhh, sizeof(rhh), 1, f);

    printf("%s: Be sure to format this drive with FORMAT (DIR=%d)\n",
           argv[0], dir);
  }
  fclose(f);
  return 0;
}



